package org.netkernel.layer0.bnf;

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;
import java.util.regex.*;

import org.netkernel.container.IMessages;


/** A group of parts within a grammar
 * @author Tony Butterfield
 */
abstract class GroupImpl implements IGrammarGroup
{
	private final String mName; 
	private final IPartEncoding mEncoding;
	private IGrammarGroup mParent;
	
	private Pattern mInnerRegex;
	private AtomicReference<Matcher> mInnerMatcher;
	
	// these values are initialised by close()
	private String mRegexString;
	private Pattern mRegex;
	private AtomicReference<Matcher> mMatcher;
	private boolean mNeedsToMatchDeeper;
	
	protected IGrammarPart[] mPartsArray;
	private int mIndexIntoParent;
	
	// these values are not used and nulled out after close2()
	private List mPartsList;
	protected NamedGroups mGroups;
	
	public GroupImpl(IGrammarGroup aParent, String aName, IPartEncoding aEncoding)
	{	mParent=aParent;
		mName=aName;
		mEncoding=aEncoding;
	}
	
	protected void setInnerRegex(String aRegex)
	{	mInnerRegex=BNFGrammarFactory.patternForString(aRegex);
		mInnerMatcher = new AtomicReference<Matcher>(mInnerRegex.matcher(""));
	}
	
	protected boolean hasInnerRegex()
	{	return mInnerMatcher!=null;
	}
	
	protected String[][] findMatchingGroups(String aIdentifier)
	{	ArrayList l=new ArrayList(16);
		Matcher m = mInnerMatcher.getAndSet(null);
		boolean restore=(m!=null);
		if (restore)
		{	m.reset(aIdentifier);
		}
		else
		{	m = mInnerRegex.matcher(aIdentifier);
		}
		while(m.find())
		{	String[] groups = new String[m.groupCount()];
			for (int i=0; i<groups.length; i++)
			{	groups[i]=m.group(i+1);
			}
			l.add(groups);
		}
		if (restore)
		{	mInnerMatcher.getAndSet(m);
		}
		String[][] result = new String[l.size()][];
		for (int i=0; i<result.length; i++)
		{	result[i]=(String[])l.get(i);
		}
		return result; 
	}
	
	public boolean needsToMatchDeeper()
	{	return mNeedsToMatchDeeper;
	}
	
	public void addPart(IGrammarPart aPart)
	{	if (mPartsList==null)
		{	mPartsList = new ArrayList();
		}
		mPartsList.add(aPart);
	}
	
	public void setParts(IGrammarPart[] aParts)
	{	mPartsArray=aParts;
	}
	
	public void close(IMessages aMessages) throws GrammarParseException
	{	
		if (mPartsList!=null)
		{	mPartsArray = new IGrammarPart[mPartsList.size()];
			for (int i=0; i<mPartsArray.length; i++)
			{	mPartsArray[i]=(IGrammarPart)mPartsList.get(i);
				if (mPartsArray[i] instanceof IGrammarGroup)
				{	((IGrammarGroup)mPartsArray[i]).close(aMessages);
				}
			}
			for (int i=0; i<mPartsArray.length; i++)
			{	mPartsArray[i]=(IGrammarPart)mPartsList.get(i);
				if (mPartsArray[i] instanceof IGrammarGroup)
				{	mNeedsToMatchDeeper|=((IGrammarGroup)mPartsArray[i]).needsToMatchDeeper();
				}
			}
			mPartsList=null;
		}
		else if (mPartsArray==null)
		{	mPartsArray = new IGrammarPart[0];
		}
		for (int i=0; i<mPartsArray.length; i++)
		{	if (mPartsArray[i] instanceof IGrammarGroup)
			{	((IGrammarGroup)mPartsArray[i]).close(aMessages);
			}
		}

		StringWriter sw=new StringWriter(128);
		mGroups = new NamedGroups();
		buildRegex(sw);
		mRegexString=sw.toString();
		boolean isStatic=mPartsArray.length==1 && mPartsArray[0] instanceof StaticPart;
		if (!isStatic)
		{	mRegex=BNFGrammarFactory.patternForString(mRegexString);
			mRegexString=mRegex.pattern();
			mMatcher=new AtomicReference<Matcher>(mRegex.matcher(""));
		}
	}
	

	
	public void close2()
	{
		mGroups=null;
		for (int i=0; i<mPartsArray.length; i++)
		{	if (mPartsArray[i] instanceof IGrammarGroup)
			{	((IGrammarGroup)mPartsArray[i]).close2();
			}
		}
	}
	
	protected abstract void buildRegex(StringWriter aSW);
	
	void setParent(GroupImpl aParent)
	{	mParent=aParent;
	}
	
	public IGrammarGroup getParent()
	{	return mParent;
	}


	public String getName()
	{	return mName;
	}
	
	public String[] matchesForGroups(String aIdentifier)
	{
		String[] result;
		if (mRegex!=null)
		{
			Matcher m = mMatcher.getAndSet(null);
			boolean restore=(m!=null);
			if (restore)
			{	m.reset(aIdentifier);
			}
			else
			{	m = mRegex.matcher(aIdentifier);
			}
			
			if (m.matches())
			{	result = new String[m.groupCount()+1];
				for (int i=0; i<result.length; i++)
				{	result[i]=m.group(i);
				}
			}
			else
			{	result=null;
			}
			
			if (restore)
			{	mMatcher.getAndSet(m);
			}
		}
		else
		{	//static group
			String comparison=((StaticPart)mPartsArray[0]).getContent();
			if (aIdentifier.equals(comparison))
			{	result = new String[] {aIdentifier,aIdentifier};
			}
			else
			{	result=null;
			}
		}
		return result;
	}
	
	public boolean match(String aIdentifier)
	{	boolean result;
		if (mRegex!=null)
		{	Matcher m = mMatcher.getAndSet(null);
			boolean restore=(m!=null);
			if (restore)
			{	m.reset(aIdentifier);
			}
			else
			{	m = mRegex.matcher(aIdentifier);
			}		
			
			result = m.matches();
			
			if (restore)
			{	mMatcher.getAndSet(m);
			}
		}
		else
		{	String comparison=((StaticPart)mPartsArray[0]).getContent();
			result = aIdentifier.equals(comparison);
		}
		return result;
	}


	public IGrammarPart getPart(int i)
	{	return mPartsArray[i];
	}
	
	public int getPartCount()
	{	return mPartsArray.length;
	}
	
	public String getRegex()
	{	return mRegexString;
	}
	
	public IPartEncoding getEncoding()
	{	return mEncoding;
	}
	
	public void writeRegex(StringWriter aSW, NamedGroups aGroups)
	{	aSW.write(getRegex());
		mIndexIntoParent=aGroups.size();
		mGroups.propagateGroups(aGroups);
	}
	
	public int getCapturedGroupIndexInParent()
	{	return mIndexIntoParent;
	}

	public boolean matches(String[] aCapturedGroups, int aOffset)
	{	boolean result=true;
	
		for (int i=0; i<mPartsArray.length; i++)
		{
			IGrammarPart part = mPartsArray[i];
			if (part instanceof IGrammarGroup)
			{	IGrammarGroup nestedGroup = (IGrammarGroup)part;
				if (nestedGroup.needsToMatchDeeper())
				{	result = nestedGroup.matches(aCapturedGroups, aOffset+nestedGroup.getCapturedGroupIndexInParent());
				}
			}
		}
		
		return result;	
	}	
	
	protected static void appendQuantifier(StringWriter aSW, int aMin, int aMax)
	{	
		if (aMin==1 && aMax==1)
		{
		}
		else	if (aMin==0 && aMax==1)
		{	aSW.write('?');
		}
		else if (aMin==0 && aMax==Integer.MAX_VALUE)
		{	aSW.write('*');
		}
		else if (aMin==1 && aMax==Integer.MAX_VALUE)
		{	aSW.write('+');
		}
		else if (aMin==aMax && aMin<Integer.MAX_VALUE)
		{	aSW.write('{');
			aSW.write(Integer.toString(aMin));
			aSW.write('}');
		}
		else if (aMin>1 && aMax==Integer.MAX_VALUE)
		{	aSW.write('{');
			aSW.write(Integer.toString(aMin));
			aSW.write(",}");
		}
		else if (aMax<Integer.MAX_VALUE)
		{	aSW.write('{');
			aSW.write(Integer.toString(aMin));
			aSW.write(',');
			aSW.write(Integer.toString(aMax));
			aSW.write('}');
		}
	}

	public String toString()
	{	String cn=this.getClass().getName();
		int i=cn.lastIndexOf('.');
		if (i>0)
		{	cn=cn.substring(i+1);
		}
		return cn+"[ name=\""+getName()+"\" regex=\""+getRegex()+"\"]";
	}
}
