package org.netkernel.layer0.representation.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.*;
import javax.xml.parsers.*;

import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.representation.IHDSInversion;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.util.Base64;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.XMLUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

public class HDSFactory
{
	public static final String HDS_LIST_DOCUMENT_ELEMENT="hds";
	public static final String HDS_NAMESPACE_URI="http://netkernel.org/hds";
	
	public static IHDSNode parseXML(InputStream aStream) throws SAXException, ParserConfigurationException, IOException
	{
		SAXParserFactory spf =  SAXParserFactory.newInstance();
		spf.setNamespaceAware(true);
		SAXParser parser=spf.newSAXParser();
		HDSSaxHandler h = new HDSSaxHandler();
		parser.parse(aStream,h);
		HDSBuilder b = new HDSBuilder();
		b.importNode(h.getRoot());
		return b.getRoot();
	}
	
	public static IHDSNode parseDOM(Node aNode) throws Exception
	{	HDSBuilder b = new HDSBuilder();
		recurseParseDOM(aNode,b);
		return b.getRoot();
	}
	
	private static void recurseParseDOM(Node aNode, HDSBuilder aBuilder) throws Exception
	{	
		switch (aNode.getNodeType())
		{	case Node.ELEMENT_NODE:
			{	String name=aNode.getNodeName();
				String text=XMLUtils.getText(aNode,true,false);
				if (text.length()==0)
				{	aBuilder.pushNode(name);
				}
				else
				{	aBuilder.pushNode(name,text);
				}
				Element e=(Element)aNode;
				NamedNodeMap atts=e.getAttributes();
				for (int i=0; i<atts.getLength(); i++)
				{	Node att=atts.item(i);
					if (HDS_NAMESPACE_URI.equals(att.getNamespaceURI()))
					{	String type=att.getNodeValue();
						Object value=parseHDSType(text,type);
						aBuilder.setValue(value);
					}
					else if (att.getNodeValue().equals(HDS_NAMESPACE_URI) && att.getNodeName().startsWith("xmlns:"))
					{	//silently ignore this namespace declaration
					}
					else
					{	aBuilder.addNode("@"+att.getNodeName(), att.getNodeValue());
					}
					
				}
				Element child=XMLUtils.getFirstChildElement(aNode);
				while(child!=null)
				{	recurseParseDOM(child, aBuilder);
					child=XMLUtils.getNextSiblingElement(child);
				}
				
				aBuilder.popNode();
			}
			break;
			case Node.DOCUMENT_NODE:
			{	Element e = ((Document)aNode).getDocumentElement();
				if(e.getNodeName().equals(HDS_LIST_DOCUMENT_ELEMENT))
				{	e=XMLUtils.getFirstChildElement(e);	//Flatten to an HDS list
					recurseParseDOM(e,aBuilder);
					while((e=XMLUtils.getNextSiblingElement(e))!=null)
					{	recurseParseDOM(e,aBuilder);					
					}
				}
				else
				{	recurseParseDOM(e,aBuilder);
				}
			}
			break;
		}
	}
	
	private static class HandlerState
	{
		private StringBuilder sb;
		private List mChildren;
		private String mType;
		private PairList mPendingPrefixMappings;
		
		private static final IHDSNode[] sNoChildren = new IHDSNode[0];
		
		public void addPrefixMapping(String aPrefixName, String aURI)
		{
			if (mPendingPrefixMappings==null)
			{	mPendingPrefixMappings = new PairList(2);
			}
			mPendingPrefixMappings.put(aPrefixName, aURI);
		}
		
		public PairList getPrefixMappings()
		{	return mPendingPrefixMappings;
		}
		
		public IHDSNode[] getChildren()
		{	IHDSNode[] result;
			if (mChildren!=null)
			{	int l=mChildren.size();
				result = new IHDSNode[l];
				for (int i=0; i<l; i++)
				{	result[i]=(IHDSNode)mChildren.get(i);
				}
			}
			else
			{	result = sNoChildren;
			}
			return result;
		}
		
		public Object getValue() throws Exception
		{	Object result=null;
			if (sb!=null)
			{	String s=sb.toString();
				if (mType!=null)
				{	result=parseHDSType(s, mType);
				}
				else
				{	result=s;
				}
			}
			return result;
		}
		
		
		public void addChild(IHDSNode aChild)
		{	if (mChildren==null)
			{	mChildren=new ArrayList(8);
			}
			mChildren.add(aChild);
		}
		
		public void addText(char[] ch, int start, int length)
		{	if (sb==null)
			{	sb=new StringBuilder(64);
			}
			
			sb.append(ch, start, length);
		}
		
		public void setType(String aType)
		{	mType=aType;
		}
	}
	
	private static class HDSSaxHandler extends DefaultHandler
	{
		private HDSNodeImpl mRoot;
		private HDSNodeImpl mCursorNode;
		private List mHandlerState=new ArrayList();
		private HandlerState mCursorState = new HandlerState();
		
		public IHDSNode getRoot()
		{	return mRoot;
		}
		
		
		
		@Override
		public void startPrefixMapping(String aPrefixName, String aURI) throws SAXException
		{	if (!aURI.equals(HDS_NAMESPACE_URI))
			{	mCursorState.addPrefixMapping(aPrefixName, aURI);
			}
		}

		public void startElement(String uri, String localName, String qName, Attributes attributes)
		{	String name=qName;
			if (mRoot==null)
			{	mCursorNode = new HDSNodeImpl(name,null);
				mRoot=mCursorNode;
			}
			else
			{	mCursorNode = new HDSNodeImpl(name,mCursorNode);
				mCursorState.addChild(mCursorNode);
			}
			
			PairList prefixMappings = mCursorState!=null?mCursorState.getPrefixMappings():null;
			
			mCursorState = new HandlerState();
			
			if (prefixMappings!=null)
			{	for (int i=0; i<prefixMappings.size(); i++)
				{	String prefix=(String)prefixMappings.getValue1(i);
					String uriValue=(String)prefixMappings.getValue2(i);
					mCursorState.addChild(new HDSNodeImpl("@xmlns:"+prefix,uriValue,mCursorNode));
				}
			}
			
			int l=attributes.getLength();
			for (int i=0; i<l; i++)
			{	name=attributes.getQName(i);
				String text=attributes.getValue(i);
				String ns=attributes.getURI(i);
				
				if (ns.equals(HDS_NAMESPACE_URI))
				{	mCursorState.setType(text);
				}
				else if (text.equals(HDS_NAMESPACE_URI) && name.startsWith("xmlns:"))
				{	//ignore hds namespace declarations
				}
				else
				{	mCursorState.addChild(new HDSNodeImpl("@"+name,text,mCursorNode));
				}
			}

			mHandlerState.add(mCursorState);
		}
		
		public void endElement(String uri, String localName, String qName) throws SAXException
		{	mCursorNode.setChildren(mCursorState.getChildren());
			try
			{	mCursorNode.setValue(mCursorState.getValue());
			}
			catch (Exception e)
			{	throw new SAXParseException("Failed to recreate value for node",null,e);
			}
			mCursorNode = (HDSNodeImpl)mCursorNode.getParent();
			int i=mHandlerState.size();
			mHandlerState.remove(i-1);
			if (i>=2)
			{	mCursorState=(HandlerState)mHandlerState.get(i-2);
			}
		}
		
		public void characters(char[] ch, int start, int length)
		{	
			boolean isWhitespace=true;
			for (int i=start+length-1; i>=start; i--)
			{	char c=ch[i];
				if (c!=' ' && c!='\n' && c!='\r' && c!='\t')
				{	isWhitespace=false;
					break;
				}
			}
			if (!isWhitespace)
			{	mCursorState.addText(ch, start, length);
			}
		}
		
	}
	
	public static Document toDOM(IHDSNode aRoot) throws NKFException, ParserConfigurationException
	{	return toDOM(aRoot,true);
	}
	
	public static Document toDOM(IHDSNode aRoot, boolean aToString) throws NKFException, ParserConfigurationException
	{	Document d=XMLUtils.newDocument();
		IHDSNode root=aRoot.getRoot();
		IHDSNode n=null;
		Element e=null;
		if(root==aRoot)
		{	//Build from root
			IHDSNode[] children=aRoot.getChildren();
			if(children!=null)
			{	if(children.length>1)
				{	//This is a flattened HDS list - create a default hds document element
					e=d.createElement(HDS_LIST_DOCUMENT_ELEMENT);
					n=root;
				}
				else if (children.length==0)
				{	e=d.createElement(HDS_LIST_DOCUMENT_ELEMENT);
					n=null;
				}
				else
				{	n=children[0];
				}
			}
		}
		else
		{	//Build from fragment
			n=aRoot;
		}
		if(e==null)
		{	e=d.createElement(n.getName());
		}
		if (!aToString)
		{	e.setAttribute("xmlns:hds", HDS_NAMESPACE_URI);
		}
		d.appendChild(e);

		if (n!=null)
		{	Object value=n.getValue();
			if(value!=null)
			{	if (aToString)
				{	Node t=d.createTextNode(value.toString());
					e.appendChild(t);
				}
				else
				{	String[] v=serialiseHDSType(value);
					e.setAttributeNS(HDS_NAMESPACE_URI, "hds:type", v[1]);
					Node t=d.createTextNode(v[0]);
					e.appendChild(t);
				}
			}
			NSLookup lookup = new NSLookup();
			recurseHDS(d, e, n, lookup,false, aToString);
		}
		return d;
	}
	
	public static Document toDOM(IHDSNodeList aList, boolean aToString) throws NKFException, ParserConfigurationException
	{
		Document d=XMLUtils.newDocument();
		Element e=d.createElement(HDS_LIST_DOCUMENT_ELEMENT);
		if (!aToString)
		{	e.setAttribute("xmlns:hds", HDS_NAMESPACE_URI);
		}
		d.appendChild(e);
		NSLookup lookup = new NSLookup();
		for (IHDSNode n: aList)
		{	recurseHDS(d, e, n, lookup,true, aToString);
		}
		return d;
	}
	
	public static IHDSInversion toHDSInversion(IHDSNode aNode)
	{	return new HDSInversionImpl(aNode);
	}
	
	
	private static void recurseHDS(Document d, Element e, IHDSNode n, NSLookup aLookup, boolean aAddThis, boolean aToString) throws NKFException
	{	
		Element e2=e;
		if (aAddThis)
		{
			String name=n.getName();
			Object value=n.getValue();
			if (name==null)
			{	throw new NKFException("bad node with null name");
			}
			else if(name.startsWith("@"))
			{	//Attribute
				if(value==null)
				{	value=""; 	//make this an empty attribute						
				}
				addAttribute(d, e, name.substring(1), value.toString(), aLookup);
			}
			else
			{	//Element
				e2=createElement(d,e,name,aLookup);
				if(value!=null)
				{	if (aToString)
					{	
						String v=value.toString();
						Node t=createNodeFor(v,d);
						e2.appendChild(t);
					}
					else
					{	String[] v=serialiseHDSType(value);
						Node t;
						if (v[1].equals(HDSType.STRING.name()))
						{	t=createNodeFor(v[0], d);
						}
						else
						{	e2.setAttributeNS(HDS_NAMESPACE_URI, "hds:type", v[1]);
							t=d.createTextNode(v[0]);
						}
						e2.appendChild(t);
					}
				}
				e.appendChild(e2);
			}
		}
		
		IHDSNode[] children=n.getChildren();
		if(children!=null)
		{	for(int i=0; i<children.length; i++)
			{	IHDSNode n2=children[i];
				recurseHDS(d, e2, n2, aLookup,true, aToString);
			}
		}
	}
	
	private static Node createNodeFor(String v, Document d)
	{	Node t;
		if (v.startsWith("<!--") && v.endsWith("-->") && v.length()>7)
		{	t=d.createComment(v.substring(4,v.length()-3));
		}
		else if (v.startsWith("<![CDATA[") && v.endsWith("]]>"))
		{	t=d.createCDATASection(v.substring(9,v.length()-3));
		}
		else
		{	t=d.createTextNode(v);
		}
		return t;
	}
	
	private static Element createElement(Document aDocument, Element aParent, String aName, NSLookup aLookup)
	{
		int i=aName.indexOf(':');
		if (i>=0)
		{	String prefix=aName.substring(0,i);
			String ns=aLookup.getURL(aParent, prefix);
			return aDocument.createElementNS(ns, aName);
		}
		else
		{	return aDocument.createElement(aName);
		}
	}
	
	private static void addAttribute(Document aDocument, Element aParent, String aName, String aValue, NSLookup aLookup)
	{
		int i=aName.indexOf(':');
		if (i>=0)
		{	String prefix=aName.substring(0,i);

			if (prefix.equals("xmlns"))
			{	String nsdecl = aName.substring(i+1);
				aLookup.putNS(aParent,nsdecl,aValue);
				aParent.setAttribute(aName,aValue);
			}
			else
			{	String ns=aLookup.getURL(aParent, prefix);
				aParent.setAttributeNS(ns, aName, aValue);
			}
		}
		else
		{	aParent.setAttribute(aName,aValue);
		}
	}

	private static class NSLookup
	{
		private TreeMap<String,String> mNamespaces = new TreeMap<String,String>();
		
		public void putNS(Element aElement, String aPrefix, String aURL)
		{	String key=aPrefix+":"+XMLUtils.getPathFor(aElement);
			mNamespaces.put(key, aURL);
		}
		
		public String getURL(Element aElement, String aPrefix)
		{	
			String result=null;
			String key=aPrefix+":"+XMLUtils.getPathFor(aElement);
			result=mNamespaces.get(key);
			if (result==null)
			{	SortedMap<String,String> m=mNamespaces.headMap(key);
				if (!m.isEmpty())
				{	String possibleMatch = m.lastKey();
					if (key.startsWith(possibleMatch))
					{	result=m.get(possibleMatch);
					}
				}
			}
			return result;			
		}
	}
	
	private enum HDSType { BYTE, SHORT, INTEGER, LONG, FLOAT, DOUBLE, BOOLEAN, CHAR, STRING, DATE, OTHER };

	
	private static Object parseHDSType(String aText, String aType) throws Exception
	{	
		//TODO
		HDSType lt;
		try
		{	lt=HDSType.valueOf(aType.toUpperCase());
		}
		catch (IllegalArgumentException e2)
		{	lt=HDSType.OTHER;
		}
		Object result;
		switch (lt)
		{	case BYTE:
				result=Byte.parseByte(aText);
				break;
			case SHORT:
				result=Short.parseShort(aText);
				break;
			case INTEGER:
				result=Integer.parseInt(aText);
				break;
			case LONG:
				result=Long.parseLong(aText);
				break;
			case FLOAT:
				result=Float.parseFloat(aText);
				break;
			case DOUBLE:
				result=Double.parseDouble(aText);
				break;
			case BOOLEAN:
				result=Boolean.parseBoolean(aText);
				break;
			case CHAR:
				if (aText.length()==1)
				{	result=aText.charAt(0);
				}
				else
				{	throw new Exception("char value expected to be single character");
				}
				break;
			case STRING:
				result=aText;
				break;
			case DATE:
				long time=Long.parseLong(aText);
				result = new Date(time);
				break;
			case OTHER:
				byte[] bytes=Base64.decode(aText);
				ObjectInputStream ois = new ObjectInputStream(new ByteArrayInputStream(bytes));
				result=ois.readObject();
				break;
			default:
				result=aText;
		}
		return result;
	}
	
	private static Map<Class,HDSType> sTypes=new HashMap<Class,HDSType>();
	static
	{	sTypes.put(Byte.class, HDSType.BYTE);
		sTypes.put(Short.class, HDSType.SHORT);
		sTypes.put(Integer.class, HDSType.INTEGER);
		sTypes.put(Long.class, HDSType.LONG);
		sTypes.put(Float.class, HDSType.FLOAT);
		sTypes.put(Double.class, HDSType.DOUBLE);
		sTypes.put(Boolean.class, HDSType.BOOLEAN);
		sTypes.put(Character.class, HDSType.CHAR);
		sTypes.put(String.class, HDSType.STRING);
		sTypes.put(Date.class, HDSType.DATE);
		sTypes.put(java.sql.Date.class, HDSType.DATE);
	}
	
	private static String[] serialiseHDSType(Object aValue) throws NKFException
	{
		String text;
		String typeString;
		HDSType type = sTypes.get(aValue.getClass());
		if (type==null)
		{	type=HDSType.OTHER;
		}
		
		switch(type)
		{	case OTHER:
				try
				{	ByteArrayOutputStream os = new ByteArrayOutputStream(1024);
					ObjectOutputStream oos = new ObjectOutputStream(os);
					oos.writeObject(aValue);
					oos.flush();
					text=Base64.encodeBytes(os.toByteArray(), Base64.GZIP);
					typeString=aValue.getClass().getName();
				}
				catch (IOException e)
				{	throw new NKFException("Failed to serialise ["+aValue.getClass().getName()+"]",null,e);
				}
				break;
			case DATE:
				text=Long.toString(((Date)aValue).getTime());
				typeString=type.name();
				break;
			default:
				text=aValue.toString();
				typeString=type.name();
				break;
		}
		return new String[] {text,typeString};
	}
}
